/* PeerState - Keeps track of the Peer state through connection callbacks.
   Copyright (C) 2003 Mark J. Wielaard

   This file is part of Snark.
   
   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2, or (at your option)
   any later version.
 
   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.
 
   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software Foundation,
   Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
*/

package org.klomp.snark;

import java.util.ArrayList;
import java.util.List;
import java.util.Set;
import java.util.HashSet;

class PeerState
{
  final Peer peer;
  final PeerListener listener;
  final MetaInfo metainfo;

  // Interesting and choking describes whether we are interested in or
  // are choking the other side.
  boolean interesting = false;
  boolean choking = true;

  // Interested and choked describes whether the other side is
  // interested in us or choked us.
  boolean interested = false;
  boolean choked = true;

  // Package local for use by Peer.
  long downloaded;
  long uploaded;

  BitField bitfield;

  // Package local for use by Peer.
  final PeerConnectionIn in;
  final PeerConnectionOut out;

  // Outstanding request
  Request currentRequest = null;
  private List outstandingRequests = new ArrayList();;

  private final static int PARTSIZE = 16384; // 16K

  PeerState(Peer peer, PeerListener listener, MetaInfo metainfo,
	    PeerConnectionIn in, PeerConnectionOut out)
  {
    this.peer = peer;
    this.listener = listener;
    this.metainfo = metainfo;

    this.in = in;
    this.out = out;
  }

  // NOTE Methods that inspect or change the state synchronize (on this).

  void keepAliveMessage()
  {

    // System.out.println(peer + " rcv alive");
    /* XXX - ignored */
  }

  synchronized void chokeMessage(boolean choke)
  {
    // System.out.println(peer + " rcv " + (choke ? "" : "un") + "choked");
    choked = choke;
    listener.gotChoke(peer, choke);
  }

  synchronized void interestedMessage(boolean interest)
  {
    // System.out.println(peer + " rcv " + (interest ? "" : "un")
    //		       + "interested");
    interested = interest;
    listener.gotInterest(peer, interest);
  }

  synchronized void haveMessage(int piece)
  {
    // System.out.println(peer + " rcv have(" + piece + ")");
    // Sanity check
    if (piece < 0 || piece >= metainfo.getPieces())
      {
	// XXX disconnect?
	// System.err.println("Got strange 'have: " + piece + "' message from "
	//		   + peer);
	return;
      }

    // Can happen if the other side never send a bitfield message.
    if (bitfield == null)
      bitfield = new BitField(metainfo.getPieces());

    bitfield.set(piece);
    if (listener.gotHave(peer, piece))
      setInteresting(true);
  }

  synchronized void bitfieldMessage(byte[] bitmap)
  {
    // System.out.println(peer + " rcv bitfield");
    if (bitfield != null)
      {
	// XXX - Be liberal in what you except?
	System.err.println("Got unexpected bitfield message from " + peer);
	return;
      }

    // XXX - Check for weird bitfield and disconnect?
    bitfield = new BitField(bitmap, metainfo.getPieces());
    setInteresting(listener.gotBitField(peer, bitfield));
  }

  synchronized void requestMessage(int piece, int begin, int length)
  {
    // System.out.println(peer + " rcv request("
    //		       + piece + ", " + begin + ", " + length + ") ");
    if (choking)
      {
	// System.err.println("Request received, but choking");
	return;
      }

    // Sanity check
    if (piece < 0
	|| piece >= metainfo.getPieces()
	|| begin < 0
	|| begin > metainfo.getPieceLength(piece)
	|| length <= 0
	|| length > 4*PARTSIZE)
      {
	// XXX - Protocol error -> disconnect?
	System.err.println("Got strange 'request: " + piece
			   + ", " + begin
			   + ", " + length
			   + "' message from " + peer);
	return;
      }

    byte[] pieceBytes = listener.gotRequest(peer, piece);
    if (pieceBytes == null)
      {
	// XXX - Protocol error-> diconnect?
	System.err.println("Got request for unknown piece: " + piece);
	return;
      }

    // More sanity checks
    if (begin >= pieceBytes.length || begin + length > pieceBytes.length)
      {
	// XXX - Protocol error-> disconnect?
	System.err.println("Got out of range 'request: " + piece
			   + ", " + begin
			   + ", " + length
			   + "' message from " + peer);
	return;
      }

    // System.err.println("Sending (" + piece + ", " + begin + ", "
    //		       + length + ")" + " to " + peer);
    out.sendPiece(piece, begin, length, pieceBytes);

    // Tell about last subpiece delivery.
    if (begin + length == pieceBytes.length)
      System.out.println("Send p" + piece + " " + peer);
  }

  /**
   * Called when some bytes have left the outgoing connection.
   * XXX - Should indicate whether it was a real piece or overhead.
   */
  void uploaded(int size)
  {
    uploaded += size;
    listener.uploaded(peer, size);
    //System.out.println(peer + " UL: " + uploaded/1024 + "K");
  }

  /**
   * Called when a partial piece request has been handled by
   * PeerConnectionIn.
   */
  synchronized void pieceMessage(int piece, int begin, int length)
  {
    // System.out.println(peer + " rcv piece("
    //		       + piece + ", " + begin + ", " + length + ") ");
    // Sanity check
    if (currentRequest == null)
      {
	System.err.println("Noting requested, but 'piece: " + begin + ", "
			   + length + "' received from " + peer);
	return;
      }

    // More sanity checks
    if (currentRequest.piece != piece
	|| currentRequest.off != begin
	|| currentRequest.len != length)
      {
	System.err.println("Unrequested 'piece: " + piece + ", "
			   + begin + "' received from " + peer);
	// XXX ARGH What now !?!
	downloaded = 0;
	return;
      }

    // XXX - If we want to keep track of overhead then we should add it
    // to the overhead first and only add it to downloaded when piece is
    // checked.
    downloaded += length;
    // System.out.println(peer + " piece DL:" + downloaded/1024 + "K");

    int pieceLength = metainfo.getPieceLength(piece);
    if (begin + length == pieceLength)
      {
	if (listener.gotPiece(peer, piece, currentRequest.bs))
	  ; // System.err.println("Got " + piece + ": " + peer);
	else
	  {
	    System.err.println("Got BAD " + piece + " from " + peer);
	    // XXX ARGH What now !?!
	    downloaded = 0;
	  }
      }

    // Anything more to request?
    if (outstandingRequests.size() > 0)
      currentRequest = (Request)outstandingRequests.remove(0);
    else
      {
	currentRequest = null;
	setInteresting(false);
      }

    // Request more if necessary/allowed.
    if (currentRequest != null && outstandingRequests.size() != 0 && !choked)
      {
	Request lastRequest
	  = (Request)outstandingRequests.get(outstandingRequests.size() - 1);
	pieceLength = metainfo.getPieceLength(lastRequest.piece);
	// Last part of a piece?
	if (lastRequest.off + lastRequest.len == pieceLength)
	  requestNextPiece();
	else
	  {
	    int nextPiece = lastRequest.piece;
	    int nextBegin = lastRequest.off + PARTSIZE;
	    byte[] bs = lastRequest.bs;
	    int maxLength = pieceLength - nextBegin;
	    int nextLength = maxLength > PARTSIZE ? PARTSIZE : maxLength;
	    Request req = new Request(nextPiece, bs, nextBegin, nextLength);
	    outstandingRequests.add(req);
	    out.sendRequest(req);
	  }
      }
  }

  synchronized void cancelMessage(int piece, int begin, int length)
  {
    // XXX - Ignored for now...
    // System.err.println("Warning: Ignoring cancel message from " + peer);
  }

  synchronized void unknownMessage(int type, byte[] bs)
  {
    System.err.println("Warning: Ignoring unknown message type: " + type
		       + " length: " + bs.length);
  }

  // Private helper functions. Not synchronized, but they should be
  // called with the lock on this held.

  private void requestNextPiece()
  {
    int nextPiece = listener.wantPiece(peer, bitfield);
    // System.out.println(peer + " want piece " + nextPiece);
    if (nextPiece != -1)
      {
	setInteresting(true);
	
	// Sanity check
	if (currentRequest == null && outstandingRequests.size() != 0)
	  throw new IllegalStateException
	    ("currentRequest == null, but outstandingRequests.size != 0");
	
	int piece_length = metainfo.getPieceLength(nextPiece);
	byte[] bs = new byte[piece_length];
	
	int length = Math.min(piece_length, PARTSIZE);
	Request req = new Request(nextPiece, bs, 0, length);
	out.sendRequest(req);

	if (currentRequest == null)
	  {
	    currentRequest = req;
	    // Add something more to the pipeline.
	    while (req.off + req.len != piece_length)
	      {
		int nextBegin = req.off + PARTSIZE;
		int maxLength = piece_length - nextBegin;
		int nextLength = maxLength > PARTSIZE ? PARTSIZE : maxLength;
		req = new Request(nextPiece, bs, nextBegin, nextLength);
		outstandingRequests.add(req);
		out.sendRequest(req);
	      }
	  }
	else
	  outstandingRequests.add(req);
      }
    else if (currentRequest == null && outstandingRequests.size() == 0)
      setInteresting(false); // We got everything we wanted
  }

  private void setInteresting(boolean interest)
  {
    if (interest != interesting)
      {
	interesting = interest;
	out.sendInterest(interest);
      }
  }

  synchronized void download(boolean download)
  {
    // System.out.println(peer + " dl " + download);
    if (download)
      {
	if (currentRequest != null)
	  {
	    // Try to resend the request
	    out.sendRequest(currentRequest);
	    out.sendRequests(outstandingRequests);
	  }
	else
	  requestNextPiece();
      }
    else
      {
	; // XXX - Should we cancel, or just keep what we get?
      }
  }

  synchronized void upload(boolean upload)
  {
    // System.out.println(peer + " up " + upload);
    choking = !upload;
    out.sendChoke(choking);
  }
}
